"""
Plugin lifecycle:
  To create a new plugin:
    * Plugin is __init__()ed
    * Plugin.prepare(data) gets called with the connected inputs, to let
      the plugin know what its inputs are.  This method should not do anything
      CPU intensive, it just checks the type of inputs are suitable.
    * Plugin.set_parameter(key, value) gets called, potentially many times.
    * Plugin.prerun() gets called and this should return sample output.
  When inputs have changed:
    * Call prepare() again.
  To change configuration
    * Call Plugin.set_parameter(), potentially several times.
    * Call Plugin.prerun() to generate new output
"""

import hashlib
import importlib
import importlib.metadata
import logging
import os.path
from collections.abc import MutableMapping
from typing import Dict, Iterable, List, Optional, Union

import numpy as np
import pandas as pd

from countess.core.logger import Logger
from countess.core.parameters import (
    ArrayParam,
    BaseParam,
    FileArrayParam,
    FileParam,
    FileSaveParam,
    MultiParam,
    StringParam,
)
from countess.utils.pandas import get_all_columns
from countess.utils.parallel import multiprocess_map

PRERUN_ROW_LIMIT = 100000


def get_plugin_classes():
    plugin_classes = set()
    try:
        # Python >= 3.10
        entry_points = importlib.metadata.entry_points().select(group="countess_plugins")
    except AttributeError:
        # Python < 3.10
        entry_points = importlib.metadata.entry_points()["countess_plugins"]

    for ep in entry_points:
        try:
            plugin_class = ep.load()
        except (ModuleNotFoundError, ImportError, NotImplementedError) as exc:
            logging.warning("%s could not be loaded: %s", ep, exc)
            continue

        if issubclass(plugin_class, BasePlugin):
            plugin_classes.add(plugin_class)
        else:
            # XXX how to warn about this?
            logging.warning("%s is not a valid CountESS plugin", plugin_class)
    return plugin_classes


def load_plugin(module_name, class_name):
    module = importlib.import_module(module_name)
    plugin_class = getattr(module, class_name)
    assert issubclass(plugin_class, BasePlugin)
    return plugin_class()


class BasePlugin:
    """Base class for all plugins.  Plugins exist as entrypoints, but also
    PluginManager checks that plugins subclass this class before accepting them
    as plugins."""

    name: str = ""
    description: str = ""
    additional: str = ""
    link: Optional[str] = None
    num_inputs: int = 1
    num_outputs: int = 1

    parameters: MutableMapping[str, BaseParam] = {}
    show_preview: bool = True

    @property
    def version(self) -> str:
        raise NotImplementedError(f"{self.__class__}.version")

    def __init__(self, plugin_name=None):
        # Parameters store the actual values they are set to, so we copy them
        # so that if the same plugin is used twice in a pipeline it will have
        # its own parameters.

        if plugin_name is not None:
            self.name = plugin_name

        self.parameters = dict((k, v.copy()) for k, v in self.parameters.items())

        # XXX should we allow django-esque declarations like this?
        # Code gets cleaner, Namespace gets cluttered, though.

        for key in dir(self):
            if isinstance(getattr(self, key), BaseParam):
                self.parameters[key] = getattr(self, key).copy()
                setattr(self, key, self.parameters[key])

    def add_parameter(self, name: str, param: BaseParam):
        self.parameters[name] = param.copy()
        return self.parameters[name]

    def set_parameter(self, key: str, value: Union[bool, int, float, str], base_dir: str = "."):
        param = self.parameters
        for k in key.split("."):
            if k == "_label" and hasattr(param, "label"):
                param.label = value
                return

            # XXX types are a mess here
            param = param[k]  # type: ignore
        if isinstance(param, (FileParam, FileSaveParam)) and value is not None:
            assert isinstance(value, str)
            param.value = os.path.join(base_dir, value)
        else:
            param.value = value  # type: ignore

    def get_parameters(self, base_dir="."):
        for key, parameter in self.parameters.items():
            yield from parameter.get_parameters(key, base_dir)

    def get_parameter_hash(self):
        """Build a hash of all configuration parameters"""
        h = hashlib.sha256()
        for k, v in self.parameters.items():
            h.update((k + "\0" + v.get_hash_value()).encode("utf-8"))
        return h

    def hash(self):
        """Returns a hex digest of the hash of all configuration parameters"""
        return self.get_parameter_hash().hexdigest()


class ProcessPlugin(BasePlugin):
    """A plugin which accepts data from one or more sources.  Each source is
    executed in its own thread and the results are collated through a thread-safe
    queue."""

    def prepare(self, sources: List[str], row_limit: Optional[int] = None):
        pass

    def process(self, data, source: str, logger: Logger) -> Iterable[pd.DataFrame]:
        """Called with each `data` input from `source`, yields results"""
        raise NotImplementedError(f"{self.__class__}.process")

    def finished(self, source: str, logger: Logger) -> Iterable[pd.DataFrame]:
        """Called when a `source` is finished and not able to
        send any more messages.  Can be ignored by most things."""
        # override this if you need to do anything
        return []

    def finalize(self, logger: Logger) -> Iterable[pd.DataFrame]:
        """Called when all sources are finished.  Can be
        ignored by most things.  This should reset the
        plugin to be ready for another use."""
        # override this if you need to do anything
        return []


class SimplePlugin(ProcessPlugin):
    def process(self, data, source: str, logger: Logger) -> Iterable[pd.DataFrame]:
        """Called with each `data` input from `source`, calls
        `callback` to send messages to the next plugin"""
        raise NotImplementedError(f"{self.__class__}.process")


class FileInputPlugin(BasePlugin):
    """Mixin class to indicate that this plugin can read files from local
    storage."""

    file_number = 0
    name = ""
    row_limit = None
    num_inputs = 0

    # used by the GUI file dialog
    file_types: List[tuple[str, Union[str, list[str]]]] = [("Any", "*")]
    file_params: MutableMapping[str, BaseParam] = {}

    def num_files(self) -> int:
        """return the number of 'files' which are to be loaded.  The pipeline
        will call code equivalent to
        `[ p.load_file(n, logger, row_limit) for n in range(0, p.num_files() ]`
        although potentially using threads, multiprocessing, etc."""
        raise NotImplementedError("FileInputMixin.num_files")

    def load_file(self, file_number: int, logger: Logger, row_limit: Optional[int] = None) -> Iterable:
        """Called potentially from multiple processes, see FileInputMixin.num_files()"""
        raise NotImplementedError("FileInputMixin.load_file")

    def prepare(self, sources: List[str], row_limit: Optional[int] = None):
        self.row_limit = row_limit

    def finalize(self, logger: Logger) -> Iterable:
        num_files = self.num_files()
        if num_files > 1:
            row_limit_per_file = self.row_limit // num_files if self.row_limit else None
            yield from multiprocess_map(self.load_file, range(0, num_files), logger, row_limit_per_file)
        elif num_files == 1:
            yield from self.load_file(0, logger, self.row_limit)


class PandasProcessPlugin(ProcessPlugin):
    DATAFRAME_BUFFER_SIZE = 100000

    def process(self, data: pd.DataFrame, source: str, logger: Logger) -> Iterable[pd.DataFrame]:
        raise NotImplementedError(f"{self.__class__}.process")


class PandasSimplePlugin(SimplePlugin):
    """Base class for plugins which accept and return pandas DataFrames.
    Subclassing this hides all the distracting aspects of the pipeline
    from the plugin implementor, who only needs to override process_dataframe"""

    input_columns: Dict[str, np.dtype] = {}

    def prepare(self, sources: list[str], row_limit: Optional[int] = None):
        self.input_columns = {}

    def process(self, data: pd.DataFrame, source: str, logger: Logger) -> Iterable[pd.DataFrame]:
        """Just deal with each dataframe as it comes.  PandasSimplePlugins don't care about `source`."""
        assert isinstance(data, pd.DataFrame)

        self.input_columns.update(get_all_columns(data))

        try:
            result = self.process_dataframe(data, logger)
            if result is not None:
                assert isinstance(result, pd.DataFrame)
                if len(result) > 0:
                    yield result

        except Exception as exc:  # pylint: disable=broad-exception-caught
            logger.exception(exc)

    def process_dataframe(self, dataframe: pd.DataFrame, logger: Logger) -> Optional[pd.DataFrame]:
        """Override this to process a single dataframe"""
        raise NotImplementedError(f"{self.__class__}.process_dataframe()")

    def finalize(self, logger: Logger) -> Iterable[pd.DataFrame]:
        yield from super().finalize(logger)
        for p in self.parameters.values():
            p.set_column_choices(self.input_columns.keys())


# class MapReduceFinalizePlugin(BasePlugin):
#    def map(self, data, logger: Logger) -> Iterable:
#        return []
#
#    def reduce(self, data: Iterable):
#        pass
#
#    def finalize(self, data: Iterable):
#        pass
#
#
# class PandasMapReduceFinalizePlugin(MapReduceFinalizePlugin):
#    def map(self, data, logger: Logger) -> Iterable[pd.DataFrame]:
#        return []
#
#    def reduce(self, data: Iterable[pd.DataFrame]) -> pd.DataFrame:
#        return pd.DataFrame()
#
#    def finalize(self, data: pd.DataFrame):
#        pass


class PandasProductPlugin(PandasProcessPlugin):
    """Some plugins need to have all the data from two sources presented to them,
    which is tricky in a pipelined environment.  This superclass handles the two
    input sources and calls .process_dataframes with pairs of dataframes.
    It is currently only used by JoinPlugin"""

    # XXX with process and finished being called from multiple threads now,
    # there need to be some locks put around things

    source1 = None
    source2 = None
    mem1: Optional[List] = None
    mem2: Optional[List] = None
    num_inputs = 2

    def prepare(self, sources: list[str], row_limit: Optional[int] = None):
        if len(sources) != 2:
            raise ValueError(f"{self.__class__} required exactly two inputs")
        self.source1, self.source2 = sources

        super().prepare(sources, row_limit)

        self.mem1 = []
        self.mem2 = []

    def process(self, data: pd.DataFrame, source: str, logger: Logger) -> Iterable[pd.DataFrame]:
        if source == self.source1:
            if self.mem1 is not None:
                self.mem1.append(data)
            assert self.mem2 is not None
            for val2 in self.mem2:
                df = self.process_dataframes(data, val2, logger)
                if len(df):
                    yield df

        elif source == self.source2:
            if self.mem2 is not None:
                self.mem2.append(data)
            assert self.mem1 is not None
            for val1 in self.mem1:
                df = self.process_dataframes(val1, data, logger)
                if len(df):
                    yield df

        else:
            raise ValueError(f"Unknown source {source}")

    def finished(self, source: str, logger: Logger) -> Iterable:
        if source == self.source1:
            # source1 is finished, mem2 is no longer needed
            self.source1 = None
            self.mem2 = None
        elif source == self.source2:
            # source2 is finished, mem1 is no longer needed
            self.source2 = None
            self.mem1 = None
        else:
            raise ValueError(f"Unknown source {source}")
        return []

    def finalize(self, logger: Logger) -> Iterable:
        # free up any memory taken up by memoization
        self.mem1 = None
        self.mem2 = None
        return []

    def process_dataframes(self, dataframe1: pd.DataFrame, dataframe2: pd.DataFrame, logger: Logger) -> pd.DataFrame:
        raise NotImplementedError(f"{self.__class__}.process_dataframes")


# XXX this might be excessively DRY but we'll see.
#
#                              --> PandasTransformAToXMixin --
#                             /       dataframe_to_series     \
# PandasTransformBasePlugin --                                 --> PandasTransformAToBMixin
#                             \       series_to_dataframe     /
#                              --> PandasTransformXToBMixin --
#
# A is one of Single (takes a value), Row (takes a pd.Series) or Dict (takes a dict)
# B is one of Single (returns a value), Tuple (returns a tuple) or Dict (returns a dict)
#
# Looked at from the point of view of the data it looks like:
#
#     P_T_BasePlugin       P_T_AToXMixin            P_T_AToBPlugin     P_T_XToBPlugin        PTBP
#
#                                               /-> process_row ---\
#     process_dataframe -> dataframe_to_series ---> process_value ---> series_to_dataframe -> merge
#                                               \-> process_dict --/
#
# Which probably seems overcomplicated but it also seems to work.
# Most plugins just need to pick a PandasTransform class and run with it.  A lot of the time it's
# going to be PandasTransformSingleToSinglePlugin.


class PandasTransformBasePlugin(PandasSimplePlugin):
    """Base classes for the nine (!!) PandasTransformXToXPlugin superclasses."""

    def series_to_dataframe(self, series: pd.Series) -> pd.DataFrame:
        raise NotImplementedError(f"{self.__class__}.series_to_dataframe()")

    def dataframe_to_series(self, dataframe: pd.DataFrame, logger: Logger) -> pd.Series:
        raise NotImplementedError(f"{self.__class__}.dataframe_to_series()")

    def process_dataframe(self, dataframe: pd.DataFrame, logger: Logger) -> Optional[pd.DataFrame]:
        try:
            # 1. A dataframe with duplicates in its index can't be merged back correctly
            # in Step 4, so we add in an extra RangeIndex to guarantee uniqueness,
            # and remove it again afterwards in Step 5.
            if dataframe.index.has_duplicates:
                dataframe.set_index(pd.RangeIndex(0, len(dataframe), name="__tmpidx"), append=True, inplace=True)

            # 2. the dataframe is transformed into
            # a series of results by PandasTransform...ToXMixin.dataframe_to_series,
            # which is expected to take each row of the dataframe, do something with
            # it and return a series of objects (values, tuples or dicts).
            series = self.dataframe_to_series(dataframe, logger)

            # 3. the series is expanded back out into rows by
            # PandasTransformXTo...Mixin.series_to_dataframe()
            dataframe_out = self.series_to_dataframe(series)

            # 4. The expanded result is merged back into the original dataframe
            dataframe_merged = dataframe.merge(dataframe_out, left_index=True, right_index=True)

            # 5. Remove extra RangeIndex if we added it in step 1.
            if "__tmpidx" in dataframe_merged.index.names:
                dataframe_merged.reset_index("__tmpidx", drop=True, inplace=True)

        except Exception as exc:  # pylint: disable=broad-exception-caught
            logger.exception(exc)
            return None

        return dataframe_merged


# XXX instead of just asserting the existence of the parameters should we
# actually create them in these mixins?


class PandasTransformSingleToXMixin:  # type: ignore [attr-defined]
    """Transformer which takes a single column, the name of which is specified
    in a ColumnChoiceParam called "column" """

    def process_value(self, value, logger: Logger):
        raise NotImplementedError(f"{self.__class__}.process_value()")

    def dataframe_to_series(self, dataframe: pd.DataFrame, logger: Logger) -> pd.Series:
        column_name = self.parameters["column"].value  # type: ignore [attr-defined]
        if column_name in dataframe:
            return dataframe[column_name].apply(self.process_value, logger=logger)
        elif column_name == dataframe.index.name:
            return dataframe.index.to_series().apply(self.process_value, logger=logger)
        elif column_name in dataframe.index.names:
            return dataframe.index.to_frame()[column_name].apply(self.process_value, logger=logger)
        else:
            null_values = [self.process_value(None, logger)] * len(dataframe)
            s = pd.Series(null_values, index=dataframe.index)
            return s


class PandasTransformRowToXMixin:
    """Transformer which takes an entire row. Less efficient than processing
    just a single value, but a bit more flexible."""

    def process_row(self, row: pd.Series, logger: Logger):
        raise NotImplementedError(f"{self.__class__}.process_row()")

    def dataframe_to_series(self, dataframe: pd.DataFrame, logger: Logger) -> pd.Series:
        return dataframe.apply(self.process_row, axis=1, args=(logger,))


class PandasTransformDictToXMixin:
    """Transformer which takes a row as a dictionary"""

    def dataframe_to_series(self, dataframe: pd.DataFrame, logger: Logger) -> pd.Series:
        if dataframe.index.names == [None]:
            # unnamed index,
            # XXX there is a bug in Pandas 2.1.x which prevents
            # args and kwargs getting passed through when raw=True
            # this seems to be fixed in Pandas 2.2.0.dev so
            # hopefully this lambda can be removed some day.
            # https://github.com/pandas-dev/pandas/issues/55009
            return dataframe.apply(
                lambda x: self.process_raw(x, list(dataframe.columns), logger),
                axis=1,
                raw=True,
            )

        columns = list(dataframe.index.names) + list(dataframe.columns)

        if len(dataframe.index.names) == 1:
            # single index
            values = (
                self.process_raw(list(index_and_data_tuple), columns, logger)
                for index_and_data_tuple in dataframe.itertuples(name=None)
            )
        else:
            # multiindex
            values = (
                self.process_raw(list(index_tuple) + list(data_tuple), columns, logger)
                for index_tuple, *data_tuple in dataframe.itertuples(name=None)
            )

        return pd.Series(values, index=dataframe.index)

    def process_dict(self, data, logger: Logger):
        raise NotImplementedError(f"{self.__class__}.process_dict()")

    def process_raw(self, data, columns, logger: Logger) -> pd.Series:
        return self.process_dict(dict(zip(columns, data)), logger=logger)


class PandasTransformXToSingleMixin:
    """Transformer which returns a single value, putting into the column specified
    by the StringParam called "output" """

    def __init__(self, *a, **k):
        super().__init__(*a, **k)
        assert isinstance(self.parameters["output"], StringParam)

    def series_to_dataframe(self, series: pd.Series) -> pd.DataFrame:
        return series.to_frame(name=self.parameters["output"].value)  # type: ignore [attr-defined]


class PandasTransformXToTupleMixin:
    """Transformer which returns a tuple of values, putting them into columns
    specifed by the StringParams "name" in the ArrayParam "output" """

    def __init__(self, *a, **k):
        super().__init__(*a, **k)
        assert hasattr(self, "parameters")
        assert isinstance(self.parameters["output"], ArrayParam)
        assert all(isinstance(pp["name"], StringParam) for pp in self.parameters["output"])

    def series_to_dataframe(self, series: pd.Series) -> pd.DataFrame:
        assert hasattr(self, "parameters")
        assert isinstance(self.parameters["output"], ArrayParam)
        assert all(isinstance(pp["name"], StringParam) for pp in self.parameters["output"])
        column_names = [
            pp.name.value or "Column %d" % n for n, pp in enumerate(self.parameters["output"], 1)
        ]  # type: ignore [attr-defined]

        series.dropna(inplace=True)
        data = series.tolist()
        if len(data):
            max_cols = max(len(d) for d in data)
            column_names = column_names[:max_cols]
            df = pd.DataFrame(data, columns=column_names, index=series.index)
            return df
        else:
            return pd.DataFrame()


class PandasTransformXToDictMixin:
    """Transformer which returns a dictionary of values, putting them into
    columns named after the dictionary keys."""

    def series_to_dataframe(self, series: pd.Series) -> pd.DataFrame:
        series.dropna(inplace=True)
        return pd.DataFrame(series.tolist(), index=series.index)


# Nine combinations of the six mixins!


class PandasTransformSingleToSinglePlugin(
    PandasTransformXToSingleMixin, PandasTransformSingleToXMixin, PandasTransformBasePlugin
):
    """Transformer which takes a single column and returns a single value"""

    def process_value(self, value, logger: Logger):
        raise NotImplementedError(f"{self.__class__}.process_value()")


class PandasTransformSingleToTuplePlugin(
    PandasTransformXToTupleMixin, PandasTransformSingleToXMixin, PandasTransformBasePlugin
):
    """Transformer which takes a single column and returns a tuple of values"""

    def process_value(self, value, logger: Logger) -> Optional[Iterable]:
        raise NotImplementedError(f"{self.__class__}.process_value()")


class PandasTransformSingleToDictPlugin(
    PandasTransformXToDictMixin, PandasTransformSingleToXMixin, PandasTransformBasePlugin
):
    """Transformer which takes a single column and returns a dictionary of values"""

    def process_value(self, value, logger: Logger) -> Optional[Dict]:
        raise NotImplementedError(f"{self.__class__}.process_value()")


class PandasTransformRowToSinglePlugin(
    PandasTransformXToSingleMixin, PandasTransformRowToXMixin, PandasTransformBasePlugin
):
    """Transformer which takes a whole row and returns a single value"""

    def process_row(self, row: pd.Series, logger: Logger):
        raise NotImplementedError(f"{self.__class__}.process_row()")


class PandasTransformRowToTuplePlugin(
    PandasTransformXToTupleMixin, PandasTransformRowToXMixin, PandasTransformBasePlugin
):
    """Transformer which takes a whole row and returns a tuple of values"""

    def process_row(self, row: pd.Series, logger: Logger):
        raise NotImplementedError(f"{self.__class__}.process_row()")


class PandasTransformRowToDictPlugin(
    PandasTransformXToDictMixin, PandasTransformRowToXMixin, PandasTransformBasePlugin
):
    """Transformer which takes a whole row and returns a dictionary of values"""

    def process_row(self, row: pd.Series, logger: Logger):
        raise NotImplementedError(f"{self.__class__}.process_row()")


class PandasTransformDictToSinglePlugin(
    PandasTransformXToSingleMixin, PandasTransformDictToXMixin, PandasTransformBasePlugin
):
    """Transformer which takes a whole row and returns a single value"""

    def process_dict(self, data, logger: Logger):
        raise NotImplementedError(f"{self.__class__}.process_dict()")


class PandasTransformDictToTuplePlugin(
    PandasTransformXToTupleMixin, PandasTransformDictToXMixin, PandasTransformBasePlugin
):
    """Transformer which takes a whole row and returns a tuple of values"""

    def process_dict(self, data, logger: Logger):
        raise NotImplementedError(f"{self.__class__}.process_dict()")


class PandasTransformDictToDictPlugin(
    PandasTransformXToDictMixin, PandasTransformDictToXMixin, PandasTransformBasePlugin
):
    """Transformer which takes a whole row and returns a dictionary of values"""

    def process_dict(self, data, logger: Logger):
        raise NotImplementedError(f"{self.__class__}.process_dict()")


class PandasInputPlugin(FileInputPlugin):
    """A specialization of the PandasBasePlugin to allow it to follow nothing,
    eg: come first."""

    def num_files(self):
        raise NotImplementedError(f"{self.__class__}.num_files()")

    def load_file(self, file_number: int, logger: Logger, row_limit: Optional[int] = None) -> Iterable[pd.DataFrame]:
        raise NotImplementedError(f"{self.__class__}.load_file()")


class PandasInputFilesPlugin(PandasInputPlugin):
    def __init__(self, *a, **k):
        # Add in filenames
        super().__init__(*a, **k)
        file_params = {"filename": FileParam("Filename", file_types=self.file_types)}
        file_params.update(self.file_params)

        self.parameters = dict(
            [("files", FileArrayParam("Files", MultiParam("File", file_params)))] + list(self.parameters.items())
        )

    def num_files(self):
        return len(self.parameters["files"].params)

    def load_file(self, file_number: int, logger: Logger, row_limit: Optional[int] = None) -> Iterable[pd.DataFrame]:
        assert isinstance(self.parameters["files"], ArrayParam)
        file_params = self.parameters["files"][file_number]
        yield self.read_file_to_dataframe(file_params, logger, row_limit)

    def read_file_to_dataframe(self, file_params, logger, row_limit=None) -> pd.DataFrame:
        raise NotImplementedError(f"{self.__class__}.read_file_to_dataframe")


class PandasOutputPlugin(PandasProcessPlugin):
    num_outputs = 0

    def process(self, data: pd.DataFrame, source: str, logger: Logger) -> Iterable[pd.DataFrame]:
        raise NotImplementedError(f"{self.__class__}.process")



import csv

import pytest

from countess.core.cmd import run as cmd_run

expected_output = """"thing","foo","bar","baz","qux","number","zz"
"bar",10,2,1,4,232,0.08620689655172414
"baz",11,3,2,1,565,0.0584070796460177
"qux",12,9,8,7,999,0.10810810810810811
"""


def test_command_invocation():
    cmd_run(["countess_cmd", "tests/simple.ini"])

    with open("tests/output.csv", "r") as fh:
        output = fh.read()
        assert output == expected_output



"""Logger base classes"""

import multiprocessing
import queue
import sys
import traceback
from typing import Iterable, Optional, Tuple


class Logger:
    """Logger Base Class"""

    def __init__(self):
        pass

    def log(self, level: str, message: str, detail: Optional[str] = None):
        """Log a message"""
        return None

    def progress(self, message: str = "Running", percentage: Optional[int] = None):
        """Record progress"""
        self.log("progress", message, str(percentage) if percentage else None)

    def info(self, message: str, detail: Optional[str] = None):
        """Log a message at level info"""
        self.log("info", message, detail)

    def warning(self, message: str, detail: Optional[str] = None):
        """Log a message at level warning"""
        self.log("warning", message, detail)

    def error(self, message: str, detail: Optional[str] = None):
        """Log a message at level error"""
        self.log("error", message, detail)

    def exception(self, exception: Exception):
        # Slightly odd calling to maintain compatibility with 3.9 and 3.10
        # XXX format more nicely
        message = traceback.format_exception(None, value=exception, tb=None)
        message += "\n\n" + "".join(traceback.format_tb(exception.__traceback__))
        self.error(str(exception), detail="".join(message))

    def clear(self):
        """Clear logs (if possible)"""
        return None


class ConsoleLogger(Logger):
    """A simple Logger which sends output to stdout and stderr"""

    def __init__(self, stdout=sys.stdout, stderr=sys.stderr, prefix: Optional[str] = None):
        self.stdout = stdout
        self.stderr = stderr
        self.prefix = prefix

    def log(self, level: str, message: str, detail: Optional[str] = None):
        if self.prefix:
            message = self.prefix + ": " + message
        if detail:
            message += " " + repr(detail)

        self.stderr.write(message + "\n")

    def progress(self, message: str = "Running", percentage: Optional[int] = None):
        if percentage is not None:
            self.log("progress", "%-20s [%-10s] %d%%" % (message, "*" * (percentage // 10), percentage))
        else:
            self.log("progress", "%-20s [ Running! ]" % message)


class MultiprocessLogger(Logger):
    def __init__(self):
        self.queue = multiprocessing.Queue()

    def log(self, level: str, message: str, detail: Optional[str] = None):
        self.queue.put((level, message, detail))

    def poll(self) -> Iterable[Tuple[str, str, str]]:
        try:
            while True:
                yield self.queue.get_nowait()
        except queue.Empty:
            pass

    def dump(self):
        return "\n".join("%s\t%s\t%s" % (a, b, c) for a, b, c in self.poll())



from typing import Iterable, Optional

import pandas as pd

from countess import VERSION
from countess.core.logger import Logger
from countess.core.parameters import ChoiceParam, ColumnChoiceParam, ColumnOrNoneChoiceParam
from countess.core.plugins import PandasSimplePlugin


class CorrelationPlugin(PandasSimplePlugin):
    """Correlations"""

    name = "Correlation Tool"
    description = "Measures Pearsons / Kendall / Spearman correlation of columns"
    version = VERSION
    link = "https://countess-project.github.io/CountESS/included-plugins/#correlation-tool"

    parameters = {
        "method": ChoiceParam("Method", choices=["pearson", "kendall", "spearman"]),
        "group": ColumnOrNoneChoiceParam("Group"),
        "column1": ColumnChoiceParam("Column 1"),
        "column2": ColumnChoiceParam("Column 2"),
    }
    columns: list[str] = []
    dataframes: list[pd.DataFrame] = []

    def prepare(self, sources: list[str], row_limit: Optional[int] = None):
        assert isinstance(self.parameters["group"], ColumnOrNoneChoiceParam)
        column1 = self.parameters["column1"].value
        column2 = self.parameters["column2"].value
        self.columns = [column1, column2]
        if self.parameters["group"].is_not_none():
            self.columns.append(self.parameters["group"].value)
        self.dataframes = []

    def process_dataframe(self, dataframe: pd.DataFrame, logger: Logger) -> None:
        self.dataframes.append(dataframe[self.columns])

    def finalize(self, logger: Logger) -> Iterable[pd.DataFrame]:
        assert isinstance(self.parameters["group"], ColumnOrNoneChoiceParam)
        column1 = self.parameters["column1"].value
        column2 = self.parameters["column2"].value
        groupby = None if self.parameters["group"].is_none() else self.parameters["group"].value

        method = self.parameters["method"].value

        dataframe = pd.concat(self.dataframes)
        if groupby:
            ds = dataframe.groupby(groupby)[column1]
            yield ds.corr(dataframe[column2], method=method).to_frame(name="correlation")
        else:
            corr = dataframe[column1].corr(dataframe[column2], method=method)
            yield pd.DataFrame([[corr]], columns=["correlation"])



import tkinter as tk
from functools import cache
from importlib.resources import as_file, files
from typing import Optional

# To keep the cache of bitmaps smaller, we always associate the image with
# the toplevel, not the individual widget it appears on.


@cache
def get_icon_toplevel(toplevel: tk.Toplevel, name: str) -> tk.Image:
    source = files("countess.gui").joinpath("icons").joinpath(f"{name}.gif")
    with as_file(source) as filepath:
        return tk.PhotoImage(master=toplevel, file=filepath)


def get_icon(widget: tk.Widget, name: str) -> tk.Image:
    return get_icon_toplevel(widget.winfo_toplevel(), name)


def info_button(parent: tk.Widget, *args, **kwargs) -> tk.Button:
    kwargs["image"] = get_icon(parent, "info")
    return tk.Button(parent, *args, **kwargs)


def add_button(parent: tk.Widget, *args, **kwargs) -> tk.Button:
    kwargs["image"] = get_icon(parent, "add")
    return tk.Button(parent, *args, **kwargs)


def delete_button(parent: tk.Widget, *args, **kwargs) -> tk.Button:
    kwargs["image"] = get_icon(parent, "del")
    return tk.Button(parent, *args, **kwargs)


class BooleanCheckbox(tk.Button):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def set_value(self, value: Optional[bool]):
        if value is None:
            self["image"] = None
            self["state"] = tk.DISABLED
            self["bd"] = 0
        elif value:
            self["image"] = get_icon(self, "check")
            self["state"] = tk.NORMAL
            self["bd"] = 1
        else:
            self["image"] = get_icon(self, "uncheck")
            self["state"] = tk.NORMAL
            self["bd"] = 1



import time

import pytest

from countess.core.config import write_config
from countess.gui.main import MainWindow, make_root


def test_open_nodes():
    root = make_root()

    mw = MainWindow(root, "tests/simple.ini")

    root.update()

    for node in mw.graph_wrapper.graph.traverse_nodes():
        mw.graph_wrapper.on_mousedown(node, None)
        root.update()
        time.sleep(0.1)
    time.sleep(1)

    write_config(mw.graph_wrapper.graph, "tests/simple.ini.output")



import hashlib
import os.path
import re
from typing import Any, Iterable, Mapping, Optional, Type, Union

import pandas as pd

PARAM_DIGEST_HASH = "sha256"


class BaseParam:
    """Represents the parameters which can be set on a plugin."""

    label: str = ""
    value: Any = None
    hide: bool = False
    read_only: bool = False

    def copy(self):
        """Plugins declare their parameters with instances of BaseParam, these
        need to be copied so that multiple Plugins (and multiple rows in an
        ArrayParam) can have distinct values"""
        raise NotImplementedError(f"Implement {self.__class__.__name__}.copy()")

    def set_value(self, value):
        self.value = value
        return self

    def get_parameters(self, key, base_dir="."):
        return ((key, self.value),)

    def get_hash_value(self):
        digest = hashlib.new(PARAM_DIGEST_HASH)
        digest.update(repr(self.value).encode("utf-8"))
        return digest.hexdigest()

    def set_column_choices(self, choices):
        pass


class SimpleParam(BaseParam):
    """A SimpleParam has a single value"""

    var_type: type = type(None)

    def __init__(self, label: str, value=None, read_only: bool = False):
        self.label = label
        if value is not None:
            self.value = value
        if read_only:
            self.read_only = True

    def clean_value(self, value):
        try:
            return self.var_type(value)
        except ValueError:
            return None

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        self._value = self.clean_value(value)

    @value.deleter
    def value(self):
        self._value = None

    def copy(self) -> "SimpleParam":
        return self.__class__(self.label, self.value, self.read_only)


class BooleanParam(SimpleParam):
    var_type = bool
    _value: bool = False

    def clean_value(self, value):
        if isinstance(value, str):
            if value in ("true", "True", "1"):
                return True
            if value in ("false", "False", "0"):
                return False
            raise ValueError(f"Can't convert {value} to boolean")
        return bool(value)


class IntegerParam(SimpleParam):
    var_type = int
    _value: int = 0

    def clean_value(self, value):
        if isinstance(value, str):
            return super().clean_value("".join(re.split(r"\D+", value)))
        else:
            return super().clean_value(value)


class FloatParam(SimpleParam):
    var_type = float
    _value: float = 0.0

    def clean_value(self, value):
        if isinstance(value, str):
            try:
                a, b = value.split(".", 1)
                s = re.split(r"\D+", a) + ["."] + re.split(r"\D+", b)
            except ValueError:
                s = re.split(r"\D+", value)
            return super().clean_value("".join(s))
        else:
            return super().clean_value(value)


class StringParam(SimpleParam):
    var_type = str
    _value: str = ""


class TextParam(StringParam):
    def clean_value(self, value):
        return re.sub("\n\n\n+", "\n\n", value)


class StringCharacterSetParam(StringParam):
    """A StringParam limited to characters from `character_set`.
    Call `clean_value` to get an acceptable value"""

    character_set: set[str] = set()

    def __init__(
        self,
        label: str,
        value=None,
        read_only: bool = False,
        *,
        character_set: Optional[set[str]] = None,
    ):
        super().__init__(label, value, read_only)
        if character_set is not None:
            self.character_set = character_set

    def clean_character(self, c: str):
        if c in self.character_set:
            return c
        elif c.upper() in self.character_set:
            return c.upper()
        else:
            return ""

    def clean_value(self, value: Any):
        value_str = str(value)
        x = "".join([self.clean_character(c) for c in value_str])
        return x

    def copy(self) -> "StringCharacterSetParam":
        return self.__class__(self.label, self.value, self.read_only, character_set=self.character_set)


def clean_file_types(file_types):
    # MacOS in particular is crashy if file_types is not to it's liking.
    # This leads to very confusing errors.  Better to throw an assertion
    # error here than bomb out later.  See #27.

    # XXX MacOS also doesn't seem to handle multiple file
    # extensions, eg: .csv.gz, whereas Linux can.
    # So maybe this function could accept those and censor
    # them if running MacOS.

    assert type(file_types) is list
    for ft in file_types:
        assert type(ft) in (tuple, list)
        assert type(ft[0]) is str
        assert type(ft[1]) in (tuple, list, str)
        if type(ft[1]) in (tuple, list):
            for ext in ft[1]:
                assert type(ext) is str
                assert ext == "*" or ext == "" or re.match(r"\.\w+$", ext), f"Invalid FileType Extension {ext}"

    return file_types


class FileParam(StringParam):
    """A StringParam for holding a filename.  Defaults to `read_only` because
    it really should be populated from a file dialog or simiar."""

    file_types = [("Any", "*")]

    _hash = None

    def __init__(self, label: str, value=None, read_only: bool = True, file_types=None):
        super().__init__(label, value, read_only)
        if file_types is not None:
            self.file_types = clean_file_types(file_types)

    def get_file_hash(self):
        if not self.value:
            return "0"
        try:
            with open(self.value, "rb") as file:
                try:
                    # Python 3.11
                    digest = hashlib.file_digest(file, PARAM_DIGEST_HASH)
                except AttributeError:
                    digest = hashlib.new(PARAM_DIGEST_HASH)
                    while True:
                        data = file.read()
                        if not data:
                            break
                        digest.update(data)
            return digest.hexdigest()
        except IOError:
            return "0"

    def get_parameters(self, key, base_dir="."):
        if self.value:
            relpath = os.path.relpath(self.value, base_dir)
        else:
            relpath = None

        return [(key, relpath)]

    def copy(self) -> "FileParam":
        return self.__class__(self.label, self.value, self.read_only, file_types=self.file_types)

    def get_hash_value(self) -> str:
        # For reproducability, we don't actually care about the filename, just
        # its hash.
        if self._hash is None:
            self._hash = self.get_file_hash()
        return self._hash


class FileSaveParam(StringParam):
    file_types = [("Any", "*")]

    def __init__(self, label: str, value=None, read_only: bool = False, file_types=None):
        super().__init__(label, value, read_only)
        if file_types is not None:
            self.file_types = file_types

    def clean_value(self, value: Union[str, tuple, list], file_types=None):
        try:
            return os.path.relpath(str(value))
        except ValueError:
            return value


class ChoiceParam(BaseParam):
    """A drop-down menu parameter choosing between options.
    Defaults to 'None'"""

    DEFAULT_VALUE = ""

    _value: Optional[str] = None
    _choice: Optional[int] = None
    choices: list[str] = []

    def __init__(
        self,
        label: str,
        value: Optional[str] = None,
        choices: Optional[Iterable[str]] = None,
    ):
        self.label = label
        self.value = value
        self.choices = list(choices or [])

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        if value is None:
            self._value = self.DEFAULT_VALUE
        else:
            self._value = value
        try:
            self._choice = self.choices.index(value)
        except ValueError:
            self._choice = None

    @property
    def choice(self):
        return self._choice

    @choice.setter
    def choice(self, choice):
        if choice is not None and 0 <= choice < len(self.choices):
            self._choice = choice
            self._value = self.choices[choice]
        else:
            self._choice = None
            self._value = None

    def set_choices(self, choices: Iterable[str]):
        self.choices = list(choices)
        if self.choices:
            if self._value not in self.choices:
                self._value = self.choices[0]
        else:
            self._value = self.DEFAULT_VALUE

    def copy(self) -> "ChoiceParam":
        return self.__class__(self.label, self.value, self.choices)


class DataTypeChoiceParam(ChoiceParam):
    DATA_TYPES: Mapping[str, tuple[type, Any, Type[SimpleParam]]] = {
        "string": (str, "", StringParam),
        "number": (float, 0.0, FloatParam),
        "integer": (int, 0, IntegerParam),
        "boolean": (bool, False, BooleanParam),
    }

    def __init__(self, label: str, value: Optional[str] = None, choices: Optional[Iterable[str]] = None):
        if not choices:
            choices = list(self.DATA_TYPES.keys())
        super().__init__(label, value, choices)

    def get_selected_type(self):
        if self.value is None:
            return None
        else:
            return self.DATA_TYPES[self.value][0]

    def cast_value(self, value):
        if value is None:
            return self.DATA_TYPES[self.value][1]
        else:
            return self.DATA_TYPES[self.value][0](value)

    def get_parameter(self, label: str, value=None) -> BaseParam:
        return self.DATA_TYPES[self.value][2](label, value)


class DataTypeOrNoneChoiceParam(DataTypeChoiceParam):
    DEFAULT_VALUE = "— NONE —"

    def __init__(self, label: str, value: Optional[str] = None, choices: Optional[Iterable[str]] = None):
        if not choices:
            choices = list(self.DATA_TYPES.keys()) + [self.DEFAULT_VALUE]
        super().__init__(label, value, choices)

    def get_selected_type(self):
        if self.value == self.DEFAULT_VALUE:
            return None
        else:
            return super().get_selected_type()

    def cast_value(self, value):
        if self.value == self.DEFAULT_VALUE:
            return None
        else:
            return super().cast_value(value)

    def is_none(self):
        return self.value == self.DEFAULT_VALUE

    def is_not_none(self):
        return self.value != self.DEFAULT_VALUE


def _dataframe_get_column(df: pd.DataFrame, col: str):
    if col in df.columns:
        return df[col]
    elif col == df.index.name:
        return df.index.to_series()
    elif hasattr(df.index, "names") and col in df.index.names:
        return df.index.to_frame()[col]
    else:
        raise ValueError(f"Column {col} not found")


class ColumnChoiceParam(ChoiceParam):
    """A ChoiceParam which DaskTransformPlugin knows
    it should automatically update with a list of columns"""

    def set_column_choices(self, choices):
        self.set_choices(list(choices))

    def get_column(self, df):
        return _dataframe_get_column(df, self.value)


class ColumnOrNoneChoiceParam(ColumnChoiceParam):
    DEFAULT_VALUE = "— NONE —"

    def set_choices(self, choices: Iterable[str]):
        super().set_choices([self.DEFAULT_VALUE] + list(choices))

    def is_none(self):
        return self.value == self.DEFAULT_VALUE

    def is_not_none(self):
        return self.value != self.DEFAULT_VALUE

    def get_column(self, df):
        if self.value == self.DEFAULT_VALUE:
            return None
        else:
            return _dataframe_get_column(df, self.value)


class ColumnOrIndexChoiceParam(ColumnChoiceParam):
    DEFAULT_VALUE = "— INDEX —"

    def set_choices(self, choices: Iterable[str]):
        super().set_choices([self.DEFAULT_VALUE] + list(choices))

    def is_index(self):
        return self.value == self.DEFAULT_VALUE

    def is_not_index(self):
        return self.value != self.DEFAULT_VALUE

    def get_column(self, df):
        if self.value == self.DEFAULT_VALUE:
            return df.index.to_series()
        else:
            return _dataframe_get_column(df, self.value)


class ColumnOrStringParam(ColumnChoiceParam):
    DEFAULT_VALUE = ""
    PREFIX = "— "

    def set_column_choices(self, choices):
        self.set_choices([self.PREFIX + c for c in choices])

    def get_column_name(self):
        if self.value.startswith(self.PREFIX):
            return self.value[len(self.PREFIX) :]
        return None

    def get_column(self, df):
        if self.value.startswith(self.PREFIX):
            col = self.value[len(self.PREFIX) :]
            return _dataframe_get_column(df, col)
        else:
            return None


class ColumnOrIntegerParam(ColumnOrStringParam):
    def clean_value(self, value):
        if isinstance(value, str):
            return int("".join(re.split(r"\D+", value)))
        else:
            return int(value)


class MultipleChoiceParam(ChoiceParam):
    pass


class ArrayParam(BaseParam):
    """An ArrayParam contains zero or more copies of `param`, which can be a
    SimpleParam or a MultiParam."""

    # XXX the only real use for this is as an array of MultiParams so I think
    # maybe we should have a TabularParam which combines ArrayParam and
    # MultiParam more directly."""

    params: list[BaseParam]

    def __init__(
        self,
        label: str,
        param: BaseParam,
        read_only: bool = False,
        min_size: int = 0,
        max_size: Optional[int] = None,
    ):
        self.label = label
        self.param = param
        self.read_only = read_only

        self.params = [param.copy() for n in range(0, min_size)]
        self.relabel()
        self.min_size = min_size
        self.max_size = max_size

    def add_row(self):
        # XXX probably should throw an exception instead of just ignoring
        if self.max_size is None or len(self.params) < self.max_size:
            pp = self.param.copy()
            self.params.append(pp)
            self.relabel()
            return pp
        return None

    def del_row(self, position: int):
        assert 0 <= position < len(self.params)

        self.params.pop(position)
        if len(self.params) < self.min_size:
            self.params.append(self.param.copy())
        self.relabel()

    def del_subparam(self, param: BaseParam):
        self.params.remove(param)
        self.relabel()

    def relabel(self):
        for n, param in enumerate(self.params):
            if param.label.startswith(self.param.label + " "):
                param.label = self.param.label + f" {n+1}"

    def copy(self) -> "ArrayParam":
        return self.__class__(self.label, self.param, self.read_only, self.min_size, self.max_size)

    def __len__(self):
        return len(self.params)

    def __getitem__(self, key):
        while len(self.params) <= int(key):
            self.add_row()
        return self.params[int(key)]

    def __contains__(self, item):
        return item in self.params

    def __iter__(self):
        return self.params.__iter__()

    @property
    def value(self):
        return [p.value for p in self.params]

    @value.setter
    def value(self, value):
        # if setting to a dictionary, keep only the values in order
        # and forget about the numbering.

        if isinstance(value, dict):
            values = sorted([(int(k), v) for k, v in value.items()])
            value = [v[1] for v in values]

        self.params = [self.param.copy().set_value(v) for v in value]

    @value.deleter
    def value(self):
        self.params = []

    def get_parameters(self, key, base_dir="."):
        for n, p in enumerate(self.params):
            yield from p.get_parameters(f"{key}.{n}", base_dir)

    def get_hash_value(self):
        digest = hashlib.new(PARAM_DIGEST_HASH)
        for p in self.params:
            digest.update(p.get_hash_value().encode("utf-8"))
        return digest.hexdigest()

    def set_column_choices(self, choices):
        self.param.set_column_choices(choices)
        for p in self.params:
            p.set_column_choices(choices)


class PerColumnArrayParam(ArrayParam):
    def __init__(self, *a, **k) -> None:
        super().__init__(*a, **k)
        self.read_only = True

    def get_parameters(self, key, base_dir="."):
        for n, p in enumerate(self.params):
            yield f"{key}.{n}._label", p.label
            yield from p.get_parameters(f"{key}.{n}", base_dir)

    def set_column_choices(self, choices):
        params_by_label = {p.label: p for p in self.params}
        self.params = [None] * len(choices)
        for num, name in enumerate(choices):
            label = f'"{name}"'
            if label in params_by_label:
                self.params[num] = params_by_label[label]
            else:
                self.params[num] = self.param.copy()
                self.params[num].label = label
            self.params[num].set_column_choices(choices)

    def get_column_params(self):
        for p in self.params:
            yield p.label, p


class FileArrayParam(ArrayParam):
    """FileArrayParam is an ArrayParam arranged per-file.  Using this class
    really just marks it as expecting to be populated from an open file
    dialog."""

    def find_fileparam(self):
        if isinstance(self.param, FileParam):
            return self.param
        if isinstance(self.param, MultiParam):
            for pp in self.param.params.values():
                if isinstance(pp, FileParam):
                    return pp
        raise TypeError("FileArrayParam needs a FileParam inside it")

    @property
    def file_types(self):
        return self.find_fileparam().file_types

    def add_files(self, filenames):
        # XXX slightly daft way of doing it.  It is setting the filename
        # of the 'template' param, and then copying that template to
        # make the new param.
        for filename in filenames:
            self.find_fileparam().value = filename
            self.add_row()


class MultiParam(BaseParam):
    params: Mapping[str, BaseParam] = {}

    def __init__(self, label: str, params: Mapping[str, BaseParam]):
        self.label = label
        self.params = params

    def copy(self) -> "MultiParam":
        pp = dict(((k, p.copy()) for k, p in self.params.items()))
        return self.__class__(self.label, pp)

    def __getitem__(self, key):
        return self.params[key]

    def keys(self):
        return self.params.keys()

    def values(self):
        return self.params.values()

    def items(self):
        return self.params.items()

    def __getattr__(self, name):
        try:
            return self.params[name]
        except KeyError as exc:
            raise AttributeError(name=name, obj=self) from exc

    def __contains__(self, item):
        return item in self.params

    def __iter__(self):
        return self.params.__iter__()

    @property
    def value(self):
        return dict((k, p.value) for k, p in self.params.items())

    @value.setter
    def value(self, value):
        for k, v in value.items():
            self.params[k].value = v

    @value.deleter
    def value(self):
        for p in self.params.values():
            del p.value

    def get_parameters(self, key, base_dir="."):
        for k, p in self.params.items():
            yield from p.get_parameters(f"{key}.{k}", base_dir)

    def get_hash_value(self):
        digest = hashlib.new(PARAM_DIGEST_HASH)
        for k, p in self.params.items():
            digest.update((str(k) + "\0" + p.get_hash_value()).encode("utf-8"))
        return digest.hexdigest()

    def set_column_choices(self, choices):
        for p in self.params.values():
            p.set_column_choices(choices)


class TabularMultiParam(MultiParam):
    pass



""" Helper functions which find HGVS variants from sequences """

# XXX this should probably become part of MAVE-HGVS library
# https://github.com/VariantEffect/mavehgvs since it is more
# generally applicable than just CountESS!

import re
from typing import Iterable, Optional

from fqfa.constants.iupac.protein import AA_CODES  # type: ignore
from fqfa.util.nucleotide import reverse_complement  # type: ignore
from fqfa.util.translate import translate_dna  # type: ignore
from rapidfuzz.distance.Levenshtein import opcodes as levenshtein_opcodes

# Insertions shorter than this won't be searched for, just included.
MIN_SEARCH_LENGTH = 10


def translate_aa(aa_seq: str) -> str:
    """translate a sequence of single-letter amino acid codes
    to a sequence of three-letter amindo acid codes

    >>> translate_aa("HYPERSENSITIVITIES")
    'HisTyrProGluArgSerGluAsnSerIleThrIleValIleThrIleGluSer'
    """

    try:
        return "".join(AA_CODES[x] for x in aa_seq)
    except KeyError as exc:
        raise ValueError("Invalid AA Sequence") from exc


def search_for_sequence(ref_seq: str, var_seq: str, min_search_length: int = MIN_SEARCH_LENGTH) -> str:
    """look for a copy of `var_seq` within `ref_seq` and if found return a reference
    in the form "offset1_offset2" otherwise return the `var_seq` itself.  Don't bother
    searching if the length of var_seq is less than MIN_SEARCH_LENGTH.

    >>> search_for_sequence("GGGGGGG", "CAT", 1)
    'CAT'

    >>> search_for_sequence("GGCATGG", "CAT", 1)
    '3_5'

    >>> search_for_sequence("GGATGAA", "CAT", 1)
    '3_5inv'

    >>> search_for_sequence("CATCATC", "CAT", 4)
    'CAT'
    """

    # XXX this could be extended to allow for inverted insertions like `850_900inv`,
    # repeated insertions like `A[26]` and/or complex insertion formats such as
    # `T;450_470;AGGG` but actually finding that kind of match is not simple.

    # XXX consider using something like re.match(r"(.+?)\1+$", var_seq) to search
    # for repeated sequences, but check that it doesn't go O(N!) or whatever
    # (so far on cpython 3.10 this seems fine)

    if len(var_seq) < min_search_length:
        return var_seq

    idx = ref_seq.rfind(var_seq)
    if idx >= 0:
        return f"{idx+1}_{idx+len(var_seq)}"

    inv_seq = reverse_complement(var_seq)
    idx = ref_seq.rfind(inv_seq)
    if idx >= 0:
        return f"{idx+1}_{idx+len(var_seq)}inv"

    return var_seq


def find_variant_dna(ref_seq: str, var_seq: str, offset: Optional[int] = 0) -> Iterable[str]:
    """ finds HGVS variants between DNA sequences in ref_seq and var_seq.
    https://varnomen.hgvs.org/recommendations/DNA/variant/insertion/
    Doesn't look for things like complex insertions (yet)

    https://varnomen.hgvs.org/bg-material/numbering/ is pretty clear
    that numbering of nucleotides starts at `1`.

    SUBSTITUTION OF SINGLE NUCLEOTIDES (checked against Enrich2)

    >>> list(find_variant_dna("AGAAGTAGAGG", "TGAAGTAGAGG"))
    ['1A>T']
    >>> list(find_variant_dna("AGAAGTAGAGG", "AGAAGTTGTGG"))
    ['7A>T', '9A>T']
    >>> list(find_variant_dna("AGAAGTAGAGG", "ATAAGAAGAGG"))
    ['2G>T', '6T>A']

    DUPLICATION OF NUCLEOTIDES

    (examples from https://varnomen.hgvs.org/recommendations/DNA/variant/duplication/ )

    >>> list(find_variant_dna(\
        "ATGCTTTGGTGGGAAGAAGTAGAGGACTGTTATGAAAGAGAAGATGTTCAAAAGAA", \
        "ATGCTTTGGTGGGAAGAAGTTAGAGGACTGTTATGAAAGAGAAGATGTTCAAAAGAA"))
    ['20dup']
    >>> list(find_variant_dna(\
        "ATGCTTTGGTGGGAAGAAGTAGAGGACTGTTATGAAAGAGAAGATGTTCAAAAGAA", \
        "ATGCTTTGGTGGGAAGAAGTAGATAGAGGACTGTTATGAAAGAGAAGATGTTCAAAAGAA"))
    ['20_23dup']

    (checked with Enrich2 ... it comes up with `g.6dupT` for the first
    and g.12dupA for the second, but that trailing symbol is not how it
    is in the HGVS docs.  The multi-nucleotide ones it gets very different
    answers for)

    >>> list(find_variant_dna("AGAAGTAGAGG", "AGAAGTTAGAGG"))
    ['6dup']
    >>> list(find_variant_dna("ATTGAAAAAAAATTAG", "ATTGAAAAAAAAATTAG"))
    ['12dup']
    >>> list(find_variant_dna("AGAAGTAGAGG", "AGAAGTAGATAGAGG"))
    ['6_9dup']
    >>> list(find_variant_dna("AAAACTAAAA", "AAAACTCTAAAA"))
    ['5_6dup']
    >>> list(find_variant_dna("AAAACTGAAAA", "AAAACTGCTGAAAA"))
    ['5_7dup']
    >>> list(find_variant_dna("AAAACTGAAAA", "AAAACTGTGAAAA"))
    ['6_7dup']

    It should always pick the 3'-most copy to identify as
    a duplicate

    >>> list(find_variant_dna("ATGGCCGCCAGCCAA", "ATGGCCGCCGCCAGCCAA"))
    ['7_9dup']


    INSERTION OF NUCLEOTIDES

    >>> list(find_variant_dna("AGAAGTAGAGG", "AGAAGTCAGAGG"))
    ['6_7insC']
    >>> list(find_variant_dna("AGAAGTAGAGG", "AGAAGTCATAGAGG"))
    ['6_7insCAT']

            0        1
            12345678901234
            AGAAGTAGAGG
            AGAAGTGAAAGAGG
             ^^^  ^^^

    GAA appears earlier in the sequence, but it's too short to bother 
    including as a reference.

    >>> list(find_variant_dna("AGAAGTAGAGG", "AGAAGTGAAAGAGG"))
    ['6_7insGAA']

    the sequence CTTTTTTTTT is long enough to use a reference for.

    >>> list(find_variant_dna("ACTTTTTTTTTAA", "ACTTTTTTTTTACTTTTTTTTTA"))
    ['12_13ins2_11']

    the most 3'ward copy of the sequence should be the one which is 
    referred to.

    >>> list(find_variant_dna(\
        "GGCATCATCATCATGGCATCATCATCATGGGG", \
        "GGCATCATCATCATGGCATCATCATCATGGCATCATCATCATGG"))
    ['30_31ins17_28']

    >>> list(find_variant_dna(\
        "GGCATCATCATCATGGCATCATCATCATGGGGCATCATCATCATGG", \
        "GGCATCATCATCATGGCATCATCATCATGGCATCATCATCATGGCATCATCATCATGG"))
    ['30_31ins33_44']

    (example from Q&A on https://varnomen.hgvs.org/recommendations/DNA/variant/insertion/ )
    "How should I describe the change ATCGATCGATCGATCGAGGGTCCC to
    ATCGATCGATCGATCGAATCGATCGATCGGGTCCC? The fact that the inserted sequence
    (ATCGATCGATCG) is present in the original sequence suggests it derives
    from a duplicative event."
    "The variant should be described as an insertion; g.17_18ins5_16."

            0        1         2         3
            1234567890123456789012345678901234
            ATCGATCGATCGATCGAGGGTCCC
            ATCGATCGATCGATCGAATCGATCGATCGGGTCCC
                ^^^^^^^^^^^  ^^^^^^^^^^^

    right so the duplicated part is 5_15 and its inserted between 17 and 18.
    I think the example in the Q&A is wrong.

    >>> list(find_variant_dna(\
        "ATCGATCGATCGATCGAGGGTCCC", \
        "ATCGATCGATCGATCGAATCGATCGATCGGGTCCC"))
    ['17_18ins5_15']

    >>> list(find_variant_dna("AAAAAAAAAACCCGGGGGGGGGGTTT", "AAAAAAAAAACCCAAAAAAAAAATTT"))
    ['14_23delins1_10']

    DELETION OF NUCLEOTIDES (checked against Enrich2)
    enrich2 has g.5_5del for the first one which isn't correct though

    >>> list(find_variant_dna("AGAAGTAGAGG", "AGAATAGAGG"))
    ['5del']
    >>> list(find_variant_dna("AGAAGTAGAGG", "AGAAAGAGG"))
    ['5_6del']

    INVERSIONS

    >>> list(find_variant_dna("AAACCCTTT", "AAAGGGTTT"))
    ['4_6inv']

    """

    ref_seq = ref_seq.strip().upper()[offset:]
    var_seq = var_seq.strip().upper()[offset:]

    if not re.match("[AGTCN]+$", ref_seq):
        raise ValueError("Invalid reference sequence")

    if not re.match("[AGTCN]+$", var_seq):
        raise ValueError("Invalid variant sequence")

    # Levenshtein algorithm finds the overlapping parts of our reference and
    # variant sequences.
    #
    # each element is a text substitution operation on the string of symbols.
    # offsets are python-style whereas HGVS offsets are 1-based and inclusive.
    #
    # 'delete' => delete symbols src_start:src_end
    # 'insert' => insert symbols dest_start:dest_end at src_start
    # 'replace' => replace symbols src_start:src_end with symbols dest_start:dest_end

    opcodes = list(levenshtein_opcodes(ref_seq, var_seq))

    # Sometimes, Levenshtein tries a little too hard to
    # find an "equal" operation between inserts, so this
    # looks for that specific case and fixes it.
    #
    # example: find_variant_string("g.", "ATGGTTGGTTCG", "ATGGTTGGTGGTTC")"
    # before: "g.[9_10insGG;10dup;12del]"
    # after: "g.[7_9dup;12del]"
    #
    # this code recognizes that if there's an "equal" sequence
    # followed by an "insert" of the same sequence, then we
    # can swap them, and *if* there's an "insert" before them
    # then we can reduce the complexity of the output by
    # swapping them and merging the two inserts.
    #
    # XXX can do something similar to turn aligned replace/equal/replace
    # sequences into a single whole-codon replace (delins) as per
    # https://varnomen.hgvs.org/recommendations/DNA/variant/substitution/
    # "two variants separated by one nucleotide, together affecting one
    # amino acid, should be described as a “delins”"
    #
    # example: find_variant_string("c.", "ATGTACAAA", "ATGGATAAA")
    # before: "c.[4T>G;6C>T]"
    # after: "c.[4_6delinsGAT]"

    for n in range(0, len(opcodes) - 2):
        op0, op1, op2 = opcodes[n : n + 3]
        if op0.tag == "insert" and op1.tag == "equal" and op2.tag == "insert":
            seq1 = var_seq[op1.dest_start : op1.dest_end]
            seq2 = var_seq[op2.dest_start : op2.dest_end]
            if seq1 == seq2:
                # extend the first insert and remove the
                # second insert (the following code ignores
                # 'equal's, so it's effectively a NOP)
                op0.dest_end = op1.dest_end
                op2.tag = "equal"

    for opcode in opcodes:
        src_start, src_end = opcode.src_start, opcode.src_end
        src_seq = ref_seq[src_start:src_end]
        dest_seq = var_seq[opcode.dest_start : opcode.dest_end]

        if opcode.tag == "delete":
            assert dest_seq == ""
            # 'delete' opcode maps to HGVS 'del' operation
            if len(src_seq) == 1:
                yield f"{src_start+1}del"
            else:
                yield f"{src_start+1}_{src_end}del"

        elif opcode.tag == "insert":
            assert src_seq == ""
            # 'insert' opcode maps to either an HGVS 'dup' or 'ins' operation

            if ref_seq[src_start - len(dest_seq) : src_start] == dest_seq:
                # This is a duplication of one or more symbols immediately
                # preceding this point.
                if len(dest_seq) == 1:
                    yield f"{src_start}dup"
                else:
                    yield f"{src_start - len(dest_seq) + 1}_{src_start}dup"
            else:
                inserted_sequence = search_for_sequence(ref_seq, dest_seq)
                yield f"{src_start}_{src_start+1}ins{inserted_sequence}"

        elif opcode.tag == "replace":
            # 'replace' opcode maps to either an HGVS '>' (single substitution) or
            # 'inv' (inversion) or 'delins' (delete+insert) operation.

            # XXX does not support "exception: two variants separated by one nucleotide,
            # together affecting one amino acid, should be described as a “delins”",
            # as this code has no concept of amino acid alignment.

            if len(src_seq) == 1 and len(dest_seq) == 1:
                yield f"{src_start+1}{src_seq}>{dest_seq}"
            elif len(src_seq) == len(dest_seq) and dest_seq == reverse_complement(src_seq):
                yield f"{src_start+1}_{src_end}inv"
            else:
                inserted_sequence = search_for_sequence(ref_seq, dest_seq)
                yield f"{src_start+1}_{src_end}delins{inserted_sequence}"


def find_variant_protein(ref_seq: str, var_seq: str, offset: Optional[int] = 0):
    """Find changes between two DNA sequences, expressed
    as amino acid changes per HGVS standard.

    identical sequences:

    >>> list(find_variant_protein("ATGGTTGGTTCA", "ATGGTTGGTTCA"))
    []

    this is a synonym: both GGT and GGC => Gly:

    >>> list(find_variant_protein("ATGGTTGGTTCA", "ATGGTTGGCTCA"))
    []

    a single AA substitution:

    >>> list(find_variant_protein("ATGGTTGGTTCA", "ATGGTTCCATCA"))
    ['Gly3Pro']

    a single AA deletion:

    >>> list(find_variant_protein("ATGGTTGGTTCA", "ATGGGTTCA"))
    ['Val2del']

    a single AA duplication:

    >>> list(find_variant_protein("ATGGTTGGTTCA", "ATGGTTGGTGGTTCA"))
    ['Gly3dup']

    a single AA insertion
    >>> list(find_variant_protein("ATGGTTGGTTCA", "ATGGTTGGTAAATCA"))
    ['Gly3_Ser4insLys']

    two substitutions are coded as a delins:

    >>> list(find_variant_protein("ATGGTTGGTTCA", "ATGGCTGCTTCA"))
    ['Val2_Gly3delinsAlaAla']

    TODO: this isn't quite correct according to
    https://hgvs-nomenclature.org/stable/recommendations/protein/extension/

    >>> list(find_variant_protein("ATGGTTGGTTCA", "ATGGTTGGTTCAAAACAG"))
    ['Ser4extLysGln']

    Protein calling should stop at the first Ter encountered:

    >>> list(find_variant_protein("ATGGTTGGTTCA", "ATGGTTTAGACA"))
    ['Gly3Ter']

    Offset lets you set the frame offset (0, 1 or 2, practically)

    """

    ref_seq = ref_seq.strip().upper()
    var_seq = var_seq.strip().upper()

    if not re.match("[AGTCN]+$", ref_seq):
        raise ValueError("Invalid reference sequence")

    if not re.match("[AGTCN]+$", var_seq):
        raise ValueError("Invalid variant sequence")

    ref_pro = translate_dna(ref_seq[offset:])[0]
    var_pro = translate_dna(var_seq[offset:])[0]

    # cut protein translations off at first '*' (terminator)
    if "*" in ref_pro:
        ref_pro = ref_pro[: ref_pro.find("*") + 1]
    if "*" in var_pro:
        var_pro = var_pro[: var_pro.find("*") + 1]

    def _ref(ref_offset):
        return f"{AA_CODES[ref_pro[ref_offset]]}{ref_offset+1}"

    opcodes = list(levenshtein_opcodes(ref_pro, var_pro))

    for opcode in opcodes:
        src_start, src_end = opcode.src_start, opcode.src_end
        src_pro = ref_pro[src_start:src_end]
        dest_pro = var_pro[opcode.dest_start : opcode.dest_end]

        if opcode.tag == "delete":
            assert dest_pro == ""
            if len(src_pro) == 1:
                yield f"{_ref(src_start)}del"
            else:
                yield f"{_ref(src_start)}_{_ref(src_end-1)}del"

        elif opcode.tag == "insert":
            assert src_pro == ""

            if ref_pro[src_start - len(dest_pro) : src_start] == dest_pro:
                if len(dest_pro) == 1:
                    yield f"{_ref(src_start-1)}dup"
                else:
                    yield f"{_ref(src_start-len(dest_pro))}_{_ref(src_start)}dup"
            elif src_start == len(ref_pro):
                # 'extension', not quite standards compliant
                yield f"{_ref(src_start-1)}ext{translate_aa(dest_pro)}"
            else:
                yield f"{_ref(src_start-1)}_{_ref(src_end)}ins{translate_aa(dest_pro)}"

        elif opcode.tag == "replace":
            # XXX handle extension if src_pro[-1] == '*'

            if len(src_pro) == 1 and len(dest_pro) == 1:
                yield f"{_ref(src_start)}{translate_aa(dest_pro)}"
            else:
                yield f"{_ref(src_start)}_{_ref(src_end-1)}delins{translate_aa(dest_pro)}"

            # If the variant protein terminated, stop translating now:
            if dest_pro[-1] == "*":
                return


def find_variant_string(
    prefix: str, ref_seq: str, var_seq: str, max_mutations: Optional[int] = None, offset: Optional[int] = 0
) -> str:
    """As above, but returns a single string instead of a generator

    MULTIPLE VARIATIONS

    >>> find_variant_string("g.", "GATTACA", "GATTACA")
    'g.='
    >>> find_variant_string("g.", "GATTACA", "GTTTACA")
    'g.2A>T'
    >>> find_variant_string("g.", "GATTACA", "GTTTAGA")
    'g.[2A>T;6C>G]'
    >>> find_variant_string("g.", "GATTACA", "GTTCAGA")
    'g.[2A>T;4T>C;6C>G]'

    >>> find_variant_string("g.", "ATGGTTGGTTC", "ATGGTTGGTGGTTC")
    'g.7_9dup'
    >>> find_variant_string("g.", "ATGGTTGGTTCG", "ATGGTTGGTGGTTC")
    'g.[7_9dup;12del]'
    >>> find_variant_string("g.", "ATGGTTGGTTC", "ATGGTTGGTGGTTCG")
    'g.[7_9dup;11_12insG]'

    PROTEINS

    >>> find_variant_string("p.", "ATGGTTGGTTCA", "ATGGTTGGTTCA")
    'p.='

    >>> find_variant_string("p.", "ATGGTTGGTTCA", "ATGGTTCCATCA")
    'p.Gly3Pro'

    >>> find_variant_string("p.", "ATGGTTGGTTCA", "ATGGGTTCA")
    'p.Val2del'

    >>> find_variant_string("p.", "ATGGTTGGTTCA", "ATGGTTGGTGGTTCA")
    'p.Gly3dup'

    >>> find_variant_string("p.", "ATGGTTGGTTCA", "ATGGCTGCTTCA")
    'p.Val2_Gly3delinsAlaAla'

    CHECK FOR INVALID INPUTS

    >>> find_variant_string("x.", "CAT", "CAT")
    Traceback (most recent call last):
     ...
    ValueError: Only prefix types 'g.', 'n.' and 'p.' accepted at this time

    >>> find_variant_string("g.", "HELLO", "CAT")
    Traceback (most recent call last):
     ...
    ValueError: Invalid reference sequence

    >>> find_variant_string("g.", "CAT", "HELLO")
    Traceback (most recent call last):
     ...
    ValueError: Invalid variant sequence

    CHECK FOR MAX MUTATIONS

    >>> find_variant_string("g.", "ATTACC", "GATTACA",1)
    Traceback (most recent call last):
     ...
    ValueError: Too many variations (2) in GATTACA

    """

    if prefix.endswith("g.") and not prefix.endswith("n."):
        variations = list(find_variant_dna(ref_seq, var_seq, offset))
    elif prefix.endswith("p."):
        variations = list(find_variant_protein(ref_seq, var_seq, offset))
    else:
        raise ValueError("Only prefix types 'g.', 'n.' and 'p.' accepted at this time")

    if len(variations) == 0:
        return prefix + "="

    if max_mutations is not None and len(variations) > max_mutations:
        raise ValueError("Too many variations (%d) in %s" % (len(variations), var_seq))

    if len(variations) == 1:
        return prefix + variations[0]
    else:
        return prefix + "[" + ";".join(variations) + "]"



from typing import Iterable, List

import pandas as pd

from countess import VERSION
from countess.core.logger import Logger
from countess.core.parameters import ChoiceParam, IntegerParam, PerColumnArrayParam
from countess.core.plugins import PandasProcessPlugin
from countess.utils.pandas import get_all_columns


class CollatePlugin(PandasProcessPlugin):
    """Collates, sorts and selects data"""

    name = "Collate"
    description = "Collate and sort records by column(s), taking the first N"
    version = VERSION
    link = "https://countess-project.github.io/CountESS/included-p  lugins/#collate"

    parameters = {
        "columns": PerColumnArrayParam(
            "Columns", ChoiceParam("Role", choices=["—", "Group", "Sort (Asc)", "Sort (Desc)"])
        ),
        "limit": IntegerParam("First N records", 0),
    }

    dataframes: List[pd.DataFrame]

    def prepare(self, *_):
        self.dataframes = []

    def process(self, data: pd.DataFrame, source: str, logger: Logger) -> Iterable:
        # XXX need a more general MapReduceFinalizePlugin class though.
        assert self.dataframes is not None

        self.dataframes.append(data)
        return []

    def finalize(self, logger: Logger) -> Iterable[pd.DataFrame]:
        assert isinstance(self.parameters["columns"], PerColumnArrayParam)
        assert self.dataframes

        df = pd.concat(self.dataframes)
        input_columns = get_all_columns(df).keys()
        self.parameters["columns"].set_column_choices(input_columns)
        column_parameters = list(zip(input_columns, self.parameters["columns"]))
        group_cols = [col for col, param in column_parameters if param.value == "Group"]
        sort_cols = {
            col: param.value.endswith("(Asc)") for col, param in column_parameters if param.value.startswith("Sort")
        }

        def sort_and_limit(df: pd.DataFrame) -> pd.DataFrame:
            df = df.sort_values(by=list(sort_cols.keys()), ascending=list(sort_cols.values()))
            if self.parameters["limit"].value > 0:
                df = df.head(self.parameters["limit"].value)
            return df

        try:
            if group_cols:
                df = df.groupby(group_cols, group_keys=False).apply(sort_and_limit)
            else:
                df = sort_and_limit(df)

            yield df
        except ValueError as exc:
            logger.exception(exc)



import ast
import os.path
import re
import sys
from configparser import ConfigParser

from countess.core.logger import ConsoleLogger, Logger
from countess.core.pipeline import PipelineGraph, PipelineNode
from countess.core.plugins import load_plugin


def default_progress_callback(n, a, b, s=""):
    print(f"{n:40s} {a:4d}/{b:4d} {s}")


def default_output_callback(output):
    sys.stderr.write(repr(output))


def read_config(
    filename: str,
    logger: Logger = ConsoleLogger(),
) -> PipelineGraph:
    """Reads `filenames` and returns a PipelineGraph"""

    cp = ConfigParser()
    cp.read(filename)

    base_dir = os.path.dirname(filename)

    pipeline_graph = PipelineGraph()
    nodes_by_name: dict[str, PipelineNode] = {}

    for section_name in cp.sections():
        config_dict = cp[section_name]

        if "_module" in config_dict:
            module_name = config_dict["_module"]
            class_name = config_dict["_class"]
            # XXX version = config_dict.get("_version")
            # XXX hash_digest = config_dict.get("_hash")
            plugin = load_plugin(module_name, class_name)
        else:
            plugin = None

        position_str = config_dict.get("_position")
        notes = config_dict.get("_notes")

        position = None
        if position_str:
            position_match = re.match(r"(\d+) (\d+)$", position_str)
            if position_match:
                position = (
                    int(position_match.group(1)) / 1000,
                    int(position_match.group(2)) / 1000,
                )

        sort = config_dict.get("_sort", "0 0").split()

        # XXX check version and hash_digest and emit warnings.

        config = [(key, ast.literal_eval(val), base_dir) for key, val in config_dict.items() if not key.startswith("_")]

        node = PipelineNode(
            name=section_name,
            plugin=plugin,
            config=config,
            position=position,
            notes=notes,
            sort_column=int(sort[0]),
            sort_descending=bool(int(sort[1])),
        )
        pipeline_graph.nodes.append(node)

        for key, val in config_dict.items():
            if key.startswith("_parent."):
                node.add_parent(nodes_by_name[val])

        nodes_by_name[section_name] = node

    return pipeline_graph


def write_config(pipeline_graph: PipelineGraph, filename: str):
    """Write `pipeline_graph`'s configuration out to `filename`"""

    cp = ConfigParser()
    base_dir = os.path.dirname(filename)

    node_names_seen = set()
    for node in pipeline_graph.traverse_nodes():
        while node.name in node_names_seen:
            num = 0
            if match := re.match(r"(.*?)\s+(\d+)$", node.name):
                node.name = match.group(1)
                num = int(match.group(2))
            node.name += f" {num + 1}"
        node_names_seen.add(node.name)

        cp.add_section(node.name)
        if node.plugin:
            cp[node.name].update(
                {
                    "_module": node.plugin.__module__,
                    "_class": node.plugin.__class__.__name__,
                    "_version": node.plugin.version,
                    "_hash": node.plugin.hash(),
                    "_sort": "%d %d" % (node.sort_column, 1 if node.sort_descending else 0),
                }
            )
        if node.position:
            xx, yy = node.position
            cp[node.name]["_position"] = "%d %d" % (xx * 1000, yy * 1000)
        if node.notes:
            cp[node.name]["_notes"] = node.notes
        for n, parent in enumerate(node.parent_nodes):
            cp[node.name][f"_parent.{n}"] = parent.name
        if node.config:
            for k, v, _ in node.config:
                cp[node.name][k] = repr(v)
        elif node.plugin:
            for k, v in node.plugin.get_parameters(base_dir):
                cp[node.name][k] = repr(v)

    with open(filename, "w", encoding="utf-8") as fh:
        cp.write(fh)


def export_config_graphviz(pipeline_graph: PipelineGraph, filename: str):
    with open(filename, "w", encoding="utf-8") as fh:
        fh.write("digraph {\n")
        for node in pipeline_graph.traverse_nodes():
            label = node.name.replace('"', r"\"")
            if node.child_nodes and not node.parent_nodes:
                fh.write(f'\t"{label}" [ shape="invhouse" ];\n')
            elif node.parent_nodes and not node.child_nodes:
                fh.write(f'\t"{label}" [ shape="house" ];\n')
            else:
                fh.write(f'\t"{label}" [ shape="box" ];\n')

            for child_node in node.child_nodes:
                label2 = child_node.name.replace('"', r"\"")
                fh.write(f'\t"{label}" -> "{label2}";\n')

        fh.write("}\n")
